###############################################################################
# (c) Copyright 2019-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import re, os
from GaudiKernel.SystemOfUnits import (
    GeV, )
from Moore import (
    options,
    run_reconstruction,
)
from Panoptes.alignment import (
    standalone_rich_mirror_align_reco, )
from RecoConf.rich_data_monitoring import (
    default_rich_monitoring_options,
    alignment_rich_monitoring_options,
)
from RecoConf.rich_reconstruction import (
    default_rich_reco_options, )
from RecoConf.hlt2_tracking import (
    make_hlt2_tracks_without_UT, )
from ruamel.yaml import YAML

isQMTest = 'QMTTEST_NAME' in os.environ

myName = (os.environ["QMTTEST_NAME"]
          if isQMTest else "rich1_mirror_align_task_noUT")

options.n_threads = (1 if isQMTest else 8)

noUT = ("noUT" in myName)

#if       'md' in options.conddb_tag:
#    mp = 'md'
#elif     'mu' in options.conddb_tag:
#    mp = 'mu'

rich = 'rich1'
radiator = 'Rich1Gas'

yaml = YAML()

# retrieve the current variant name
with open(rich + '_variant.yml') as inp:
    current_variant = yaml.load(inp)

if not isQMTest:
    # retrieve number of events
    em = re.findall(r'([\d]+)k', current_variant)[0]
    iem = int(em)
    if iem == 0:
        options.evt_max = 100
    else:
        options.evt_max = iem * 1000

# prepare updates:
default_reco_opts = {}
align_opts = {}

# start updating the dicts:

default_reco_opts.update({"PhotonSelection": 'None'})

align_opts.update({"Variant": current_variant})

if 'subset' in current_variant:
    central_mirr_combs = {
        "PrebookHistos": [
            ('p00', 's03'),
            ('p01', 's06'),
            ('p02', 's12'),
            ('p03', 's09'),
        ],
    }
    align_opts.update(central_mirr_combs)

if 'minp40' in current_variant:
    align_opts.update({"MinP4Align": 40. * GeV})

if 'theta2' in current_variant:
    align_opts.update({"DeltaThetaRange": 0.002})

# prepare part of the overridden options in use
# to be recorded into the YAML file
align_opts_dump = alignment_rich_monitoring_options(
    radiator=radiator, init_override_opts=align_opts)

# prepare override of the default_rich_monitoring_options
default_tight_track_sel = default_rich_monitoring_options(
)["TightTrackSelection"]
default_tight_track_sel["MinP"] = 40.0 * GeV
default_moni_opts = {
    "TightTrackSelection": default_tight_track_sel,
    "UseUT": not noUT,
}

# add overridden default_rich_monitoring_options
# to the part of the options in use to be recorded into the YAML file
align_opts_dump.update(
    default_rich_monitoring_options(init_override_opts=default_moni_opts))

# retrieve the timestamp
ts = re.findall(r'[\d]{4}-[\d]{2}-[\d]{2}T[\d]{2}-[\d]{2}', current_variant)[0]

# prepare options in use to be recorded into the YAML file
opts_dump = {}
opts_dump[ts] = {}
opts_dump[ts]['reco_opts'] = align_opts_dump

# append YAML file with the part of the options in use
with open(rich + '_reco_opts.yml', 'a') as out:
    yaml.dump(opts_dump, out)

# retrieve list of reco tasks
mirror_align_tasks = []
tasks = re.findall(r'k(_[\w]+)', current_variant)
mirror_align_tasks = re.findall(r'_([A-Za-z]+)', tasks[0])

# retrieve iteration number substring
with open(rich + '_iter_number.yml') as inp:
    iN = '_i' + str(yaml.load(inp))

if not isQMTest:
    options.histo_file = current_variant + '_histos' + iN + '.root'

with make_hlt2_tracks_without_UT.bind(use_pr_kf=True, light_reco=True, fast_reco=True),\
     default_rich_reco_options.bind(init_override_opts=default_reco_opts), \
     default_rich_monitoring_options.bind(init_override_opts=default_moni_opts), \
     alignment_rich_monitoring_options.bind(radiator=radiator, \
                                            init_override_opts=align_opts), \
     standalone_rich_mirror_align_reco.bind(RichGas=radiator, \
                                            MirrorAlignTasks=mirror_align_tasks, \
                                            noUT=noUT):
    run_reconstruction(options, standalone_rich_mirror_align_reco)
