###############################################################################
# (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
__authors__ = "Claire Prouve, Paras Naik"
__copyright__ = "University of Bristol, November 2022"

import os
import shutil
import re
import math
import datetime
import time
from time import time
from time import gmtime, strftime
import subprocess, shlex
import sys
from ROOT import TH1, TFile
from ruamel.yaml import YAML


class SetupHelper:
    def __init__(self, _alignConf, flat=False):
        self.alignConf = _alignConf
        self.combAndMirrSubsets = self.alignConf.getProp('combAndMirrSubsets')
        self.nameStr = self.alignConf.getProp('nameStr')
        self.tiltNames = self.alignConf.getProp('tiltNames')
        self.tiltNamesLength = len(self.tiltNames)
        self.workdir = self.alignConf.getProp('WorkDir')
        self.whichRich = self.alignConf.getProp('Rich')
        self.YAMLio = self.alignConf.getProp('YAMLio')
        self.useDD4hep = self.alignConf.getProp('useDD4hep')
        self.autoUpdate = self.alignConf.getProp('autoUpdate')
        self.autoUpdateNoV = self.alignConf.getProp('autoUpdateNoV')
        self.magnFactorsMode = self.alignConf.getProp('magnFactorsMode')
        self.minAverageBinPop = self.alignConf.getProp('minAverageBinPop')
        self.phiBinFactor = self.alignConf.getProp('phiBinFactor')
        self.minFracPhiBinsPopulated = self.alignConf.getProp(
            'minFracPhiBinsPopulated')
        self.deltaThetaWindow = self.alignConf.getProp('deltaThetaWindow')
        self.combinFitMethod = self.alignConf.getProp('combinFitMethod')
        self.solutionMethod = self.alignConf.getProp('solutionMethod')
        self.coeffCalibTilt = self.alignConf.getProp('coeffCalibTilt')
        self.magnifDir = self.alignConf.getProp('magnifDir')
        self.XML_YAML_configDir = self.alignConf.getProp('XML_YAML_configDir')
        self.testing = self.alignConf.getProp('testing')
        self.fixSinusoidShift = self.alignConf.getProp('fixSinusoidShift')
        self.experts = ['Paras.Naik@cern.ch']
        #        self.experts = ['Paras.Naik@cern.ch','samuel.maddrell-mander@cern.ch','matthew.george.chapman@cern.ch','claire.prouve@cern.ch']
        self.developers = ['lhcb-rich-mirror-alignment-development@cern.ch']
        self.emails = ['Paras.Naik@cern.ch']
        #        self.emails = ['lhcb-rich-mirror-alignment-development@cern.ch','jonesc@hep.phy.cam.ac.uk','antonis.papanestis@stfc.ac.uk','m.mccann@imperial.ac.uk','silvia.gambetta@cern.ch']
        self.directoryTime = strftime("%Y%m%d_%H%M%S", gmtime())
        self.now = datetime.datetime.now()
        self.dt = self.now.isoformat(
            timespec='minutes')  # Anatoly's old timestamp
        self.dt = self.dt.replace(":", "-")  # Anatoly's new timestamp
        self.savedir = self.alignConf.getProp('SaveDir') + 'Rich' + str(
            self.whichRich) + '/' + self.directoryTime + '/'
        self.dataVariant = self.alignConf.getProp('dataVariant')
        self.xmlMDCS = self.alignConf.getProp('xmlMDCS')
        self.trendDateRange = self.alignConf.getProp('trendDateRange')
        self.finalCKres = None
        self.finalCKresErr = None
        self.flat = flat

    def setupMagnifFiles(self):
        for f in os.listdir(self.magnifDir):
            if self.flat is False:
                shutil.copy(self.magnifDir + "/" + f, self.workdir)
            else:
                shutil.copyfile(self.magnifDir + "/" + f,
                                self.workdir + "/" + f)

    def startMinIt(self, m_it):
        removeIt = (m_it - m_it % self.tiltNamesLength) / self.tiltNamesLength
        files = os.listdir(self.workdir)

        re_fitFolders = re.compile(r"Rich" + str(self.whichRich) +
                                   "MirrCombinFit_" + self.nameStr +
                                   "[\w\.]*_i" + str(removeIt))
        re_alignFiles = re.compile(r"Rich" + str(self.whichRich) +
                                   "MirrAlign[a-zA-z]*_i" + str(removeIt) +
                                   "\.txt")
        fitFolders = [x for x in files if re_fitFolders.match(x)]
        alignFiles = [x for x in files if re_alignFiles.match(x)]

        if (m_it % self.tiltNamesLength == 0):
            connectStr = ""
        else:
            connectStr = "_"
        yaml = YAML()
        with open(self.workdir + '/' + 'rich' + str(self.whichRich) +
                  '_variant.yml') as finp:
            variant = yaml.load(finp)
        rootFile = variant + "_histos_" + self.tiltNames[
            m_it % self.tiltNamesLength] + "_i" + str(removeIt) + ".root"  #

        for f in fitFolders:
            shutil.rmtree(self.workdir + "/" + f)
        for f in alignFiles:
            os.remove(self.workdir + "/" + f)
        if os.path.exists(self.workdir + "/" + rootFile):
            os.remove(self.workdir + "/" + rootFile)

    def monitor(self, n_it, compareFILE, forcedUpdate_ChMagPol, flat=False):
        from AlignMonitor import AlignMonitor
        alignMonitor = AlignMonitor(self.alignConf, n_it, flat)
        alignMonitor.performMonitoring(compareFILE, forcedUpdate_ChMagPol)
        self.finalCKres = alignMonitor.finalCKres
        self.finalCKresErr = alignMonitor.finalCKresErr
        return alignMonitor.isInsane

    def finalize(self,
                 vN_DB_str,
                 conv,
                 n_it,
                 farm_config_time,
                 elapsed_runtimes,
                 DBUpdated,
                 nEventsListList,
                 edit=None,
                 flat=False):
        nEventsList = nEventsListList[0]
        nEvents = nEventsListList[1]

        summaryFile = self.workdir + "/summary.txt"
        self.writeSummary(vN_DB_str, conv, n_it, farm_config_time,
                          elapsed_runtimes, summaryFile, DBUpdated, nEvents,
                          nEventsList, flat)

        if flat is False:
            if not os.path.exists(self.savedir):
                os.makedirs(self.savedir)
        else:
            print("INFO: You are running the flat Iterator.")
            print("INFO: Cannot save in usual place (" + self.savedir + ").")
            print(
                "INFO: Instead will save in the directory you launched the flat Iterator from."
            )
            self.savedir = "./" + '/' + self.directoryTime + '/'
            if not os.path.exists(self.savedir):
                os.makedirs(self.savedir)
        for f in os.listdir(self.workdir):
            if os.path.isdir(self.workdir + "/" + f):
                shutil.copytree(self.workdir + "/" + f, self.savedir + '/' + f)
            else:
                if self.flat is False:
                    shutil.copy2(self.workdir + "/" + f, self.savedir)
                else:
                    shutil.copyfile(self.workdir + "/" + f,
                                    self.savedir + "/" + f)
        print("INFO: savedir " + self.savedir)
        # NOTE: *Any file created after this in the workdir*
        #   needs to be *manually* saved in the savedir

        # Create the trend plots automatically, now that the savedir is written
        import trendHelper
        print("INFO: Now producing the trend plots for RICH" +
              str(self.whichRich) + ".")
        # Files for the Alignment trend plots
        trendFile_Fills_hollow = self.workdir + "/" + 'Rich' + str(
            self.whichRich) + '_AlignTrend_Fills_hollow.pdf'
        trendFile_Dates_hollow = self.workdir + "/" + 'Rich' + str(
            self.whichRich) + '_AlignTrend_Dates_hollow.pdf'
        # set dates
        minDate = self.trendDateRange[
            'minDate']  # 2017 data: 20170602 / Rich2: 20170605
        maxDate = self.trendDateRange['maxDate']  # 2017 data: 20171129
        #now = datetime.datetime.now()
        if minDate is None:
            minDate = int(
                self.now.year) * 10000 + 1 * 100 + 1  # beginning of run year
        if maxDate is None:
            maxDate = int(
                self.now.year) * 10000 + 12 * 100 + 31  # end of run year

        print(
            'INFO: Now making the "hollow" trend plot with Fills on the x-axis, with minDate '
            + str(minDate) + ' and maxDate ' + str(maxDate))
        P_Fills_hollow = trendHelper.Plotter(self.whichRich, 'hollow', "Fills",
                                             minDate, maxDate,
                                             trendFile_Fills_hollow)
        alignmentsExist = P_Fills_hollow.GetAlignments()
        if alignmentsExist:
            P_Fills_hollow.calculate_tilts()

        print(
            'INFO: Now making the "hollow" trend plot with Dates on the x-axis, with minDate '
            + str(minDate) + ' and maxDate ' + str(maxDate))
        P_Dates_hollow = trendHelper.Plotter(self.whichRich, 'hollow', "Dates",
                                             minDate, maxDate,
                                             trendFile_Dates_hollow)
        alignmentsExist = P_Dates_hollow.GetAlignments()
        if alignmentsExist:
            P_Dates_hollow.calculate_tilts()

        # Identify important file names
        ChangeWRTDBFile = self.savedir + "/Rich" + str(
            self.whichRich) + '_ChangeWRTDB_' + str(n_it + 1) + '.txt'
        AlignSummaryFile = self.savedir + "/Rich" + str(
            self.
            whichRich) + '_AlignSummary.pdf'  # For RICH piquet and experts
        AlignMonitorFile = self.savedir + "/Rich" + str(
            self.whichRich) + '_AlignMonitor.pdf'  # For Monitoring
        AlignTrendFile_Fills_hollow = self.savedir + "/Rich" + str(
            self.whichRich
        ) + '_AlignTrend_Fills_hollow.pdf'  # Automatic alignment trend plot
        AlignTrendFile_Dates_hollow = self.savedir + "/Rich" + str(
            self.whichRich
        ) + '_AlignTrend_Dates_hollow.pdf'  # Automatic alignment trend plot

        AlignTestingFile = self.workdir + "/Rich" + str(
            self.whichRich
        ) + '_AlignTesting.pdf'  # Not used just yet; May not exist

        # Copy the "hollow" trend file with Fills on the x-axis from the workdir to the savedir
        # This is the only one we send to the RICH ELOG
        if os.path.exists(trendFile_Fills_hollow) and os.path.exists(
                self.savedir):
            print(
                'INFO: storing "hollow" alignment trend plot with Fills on the x-axis in file: '
                + AlignTrendFile_Fills_hollow)
            if self.flat is False:
                shutil.copy2(trendFile_Fills_hollow,
                             AlignTrendFile_Fills_hollow)
            else:
                shutil.copyfile(trendFile_Fills_hollow,
                                AlignTrendFile_Fills_hollow)
        else:
            print("INFO: " + trendFile_Fills_hollow + " or " + self.savedir +
                  " does not exist.")
        sys.stdout.flush()

        # Copy the "hollow" trend file with Dates on the x-axis from the workdir to the savedir
        # This is not sent to the RICH ELOG, but available for AlignmentView
        if os.path.exists(trendFile_Dates_hollow) and os.path.exists(
                self.savedir):
            print(
                'INFO: storing "hollow" alignment trend plot with Dates on the x-axis in file: '
                + AlignTrendFile_Dates_hollow)
            if self.flat is False:
                shutil.copy2(trendFile_Dates_hollow,
                             AlignTrendFile_Dates_hollow)
            else:
                shutil.copyfile(trendFile_Dates_hollow,
                                AlignTrendFile_Dates_hollow)
        else:
            print("INFO: " + trendFile_Dates_hollow + " or " + self.savedir +
                  " does not exist.")
        sys.stdout.flush()

        # Write files and information to the RICH and Alignment Monitor ELOGs, and also send email
        alignment_time = elapsed_runtimes['elapsed_time_run']
        self.writeInLogbook(
            vN_DB_str, conv, n_it, farm_config_time, alignment_time,
            summaryFile, ChangeWRTDBFile, AlignSummaryFile, AlignMonitorFile,
            AlignTrendFile_Fills_hollow, DBUpdated, nEvents, edit, flat)

    def writeInLogbook(self,
                       vN_DB_str,
                       conv,
                       n_it,
                       farm_config_time,
                       alignment_time,
                       summaryFile,
                       ChangeWRTDBFile,
                       AlignSummaryFile,
                       AlignMonitorFile,
                       AlignTrendFile_Fills_hollow,
                       DBUpdated,
                       nEvents,
                       edit,
                       flat=False):
        hh, r = divmod(farm_config_time, 3600)
        mm, ss = divmod(r, 60)
        farm_config_time_str = "{:0>2}:{:0>2}:{:05.2f}".format(
            int(hh), int(mm), ss)
        hh, r = divmod(alignment_time, 3600)
        mm, ss = divmod(r, 60)
        alignment_time_str = "{:0>2}:{:0>2}:{:05.2f}".format(
            int(hh), int(mm), ss)

        if flat is False:
            import OnlineEnv as Online
            FillAlignment = self.getFillNumber(Online.DeferredRuns[0])
        else:
            FillAlignment = 0

        # RICH ELOG
        host = 'lblogbook.cern.ch'
        port = 8080
        username = 'common Common\\!'
        logbook = 'RICH'
        author = 'MirrAlign Monitor'  # Or any other name
        activity = 'Alignment'
        subject = ''
        text = ''
        instruction_file = 'https://twiki.cern.ch/twiki/bin/view/LHCb/LHCbRichMirrorAlignShiftInfo'
        subjText = ''
        if DBUpdated:
            subjText += ' updated,'
        elif self.testing and DBUpdated is None:  # if testing is on, regardless of autoUpdate setting
            subjText += ' not updated due to testing = True,'
        elif DBUpdated is None and self.autoUpdate is False:  # testing and autoUpdate were off to begin with
            subjText += ' not updated due to autoUpdate = False, needs MirrAlign expert decision,'
        elif DBUpdated is None and self.autoUpdateNoV is True:  # testing off, but autoUpdate was on at the start, then switched off
            subjText += ' not updated due to sanity check issue, needs MirrAlign expert decision,'
        elif DBUpdated is None and self.autoUpdateNoV is False:  # testing off, but autoUpdate was on at the start, then switched off
            subjText += ' not updated due to sanity check issue,'  # existing version was provided
        else:
            subjText += ''
        if self.testing:
            subject += 'TEST [' + self.dataVariant + ']: '
            text += 'TEST [' + self.dataVariant + ']: This is a test alignment being run by a mirror alignment expert; RICH piquets may ignore it. \n'
        text += "The Mirror "
        text += str(activity)
        text += " for RICH"
        text += str(self.whichRich)
        if conv:
            text += " ran for "
            text += alignment_time_str
            text += " (excluding an LHCbA configuration time of "
            text += farm_config_time_str
            text += ") and converged. "
            subject += 'RICH' + str(
                self.whichRich
            ) + ' alignment ' + self.directoryTime + subjText + ' converged'
        else:
            text += " ran for "
            text += alignment_time_str
            text += " (excluding an LHCbA configuration time of "
            text += farm_config_time_str
            text += ") and did not converge. "
            subject += 'RICH' + str(
                self.whichRich
            ) + ' alignment ' + self.directoryTime + subjText + ' incomplete'
        text += "We processed approximately "
        text += str(nEvents)
        text += " events per iteration from the RICH"
        text += str(self.whichRich)
        text += "Decision of our Hlt1CalibRICHMirror HLT line. "
        text += str(n_it + 1)
        text += " iteration(s) occured. Mirror "
        text += str(activity)
        text += " monitoring information and plots are attached; shifter's instructions can be found at "
        text += str(instruction_file)
        text += " and the output directory timestamp is "
        text += self.directoryTime
        text += "."
        if DBUpdated is None:
            text += " There was a significant change from the alignment in the DB, or a change in magnet polarity." \
                  + "However, the final state of autoUpdate was set to False (even if the initial state of autoUpdate was True), " \
                  + "or the alignment is in testing mode, so the DB was not updated. " \
                  + "If the automatic update feature was disabled, a sanity check was not satisfied. " \
                  + "Please notify a MirrAlign expert to investigate whether this alignment (" + vN_DB_str + ") should be installed."
            # We should maybe provide more info here
            if self.autoUpdate and self.autoUpdateNoV:
                text += " Please note: autoUpdate was initially True, and autoUpdateNoV was set to True. " \
                      + "This means that an alignment version has NOT been provided to HLT2."
        elif DBUpdated:
            text += " Due to a significant change from the alignment in the DB, or a change in magnet polarity, the DB was automatically updated to " + vN_DB_str + "."  # We should probably provide more info here
        else:
            text += " The alignment in the DB is still " + vN_DB_str + "."
        if not self.autoUpdate:
            text += " The initial state of autoUpdate was set to False in the configuration."
        command = 'elog -h "'
        command += str(host)
        command += '" -p '
        command += str(port)
        try:
            command += ' -e ' + str(int(edit))
        except:
            pass
        command += ' -l "'
        command += str(logbook)
        command += '" -u '
        command += str(username)
        command += ' -a "Author='
        command += str(author)
        #command += '" -a "System='
        command += '" -a "System_19='  # Change 19 if Alignment changes its order in the ELOG (ask the current RICH expert)
        command += str(activity)
        command += '" -a "Subject='
        command += str(subject)
        command += '" "'
        command += str(text)
        command += '"'

        if os.path.exists(summaryFile):
            command += ' -f "{0}"'.format(summaryFile)
        if os.path.exists(ChangeWRTDBFile):
            command += ' -f "{0}"'.format(ChangeWRTDBFile)
        if os.path.exists(AlignSummaryFile):
            command += ' -f "{0}"'.format(AlignSummaryFile)
        if os.path.exists(AlignTrendFile_Fills_hollow):
            command += ' -f "{0}"'.format(AlignTrendFile_Fills_hollow)
        cmd = shlex.split(command)
        try:
            elogret = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            messageID = int(
                re.search('.*Message successfully transmitted, ID=(\d*)',
                          elogret).groups()[0])
        except:
            messageID = None

        # Alignment ELOG
        logbookAlignment = 'Alignment monitoring'  # always
        authorAlignment = 'monibot'  # always
        activityAlignment = 'Rich' + str(self.whichRich) + ' Mirrors'  # always
        TypeAlignment = 'Convergence'  # always
        #subjectAlignment = 'IGNORE-- '
        #textAlignment = 'Please Ignore: \n '
        subjectAlignment = ''
        textAlignment = ''
        StatusAlignment = 'Good'  # or can be changed below to 'Unchecked' or 'Bad' (for now 'Bad' never happens)

        if self.testing:
            subjectAlignment += 'TEST [' + self.dataVariant + ']: '
            textAlignment += 'A test mirror alignment for Rich' + str(
                self.whichRich
            ) + ' was performed for fill ' + str(
                FillAlignment
            ) + " by a RICH mirror alignment expert. No action is necessary; ignore the following. \n "
        if DBUpdated is None:
            subjectAlignment += 'No alignment update-- Needs MirrAlign expert decision'
            textAlignment += 'Rich' + str(
                self.whichRich
            ) + ' mirror alignment for fill ' + str(
                FillAlignment
            ) + " is requesting an update be made, but for some reason this was not done automatically. Please ask the Rich Mirror Alignment experts to investigate the draft alignment (" + vN_DB_str + "). Monitoring plots in the attachment; shifter's instructions can be found at: " + instruction_file
            StatusAlignment = 'Unchecked'
        elif DBUpdated:
            subjectAlignment += 'Monitoring plots'
            textAlignment += 'Updated Rich' + str(
                self.whichRich
            ) + ' mirror alignment for fill ' + str(
                FillAlignment
            ) + " (" + vN_DB_str + "), monitoring plots in the attachment; shifter's instructions can be found at: " + instruction_file
            StatusAlignment = 'Unchecked'
        else:
            subjectAlignment += 'No alignment update'
            textAlignment += 'No Rich' + str(
                self.whichRich) + ' alignment update needed for fill ' + str(
                    FillAlignment
                ) + '. The DB alignment is still ' + vN_DB_str + '.'
        if self.testing:
            StatusAlignment = 'Test'

        command = 'elog -h "'
        command += str(host)
        command += '" -p '
        command += str(port)
        command += ' -l "'
        command += str(logbookAlignment)
        command += '" -u '
        command += str(username)
        command += ' -a "Author='
        command += str(authorAlignment)
        command += '" -a "System='
        command += str(activityAlignment)
        command += '" -a "Subject='
        command += str(subjectAlignment)
        command += '" -a "Type='
        command += str(TypeAlignment)
        command += '" -a "Fill='
        command += str(FillAlignment)
        command += '" -a "Status='
        command += str(StatusAlignment)
        command += '" "'
        command += str(textAlignment)
        command += '"'

        if DBUpdated or DBUpdated is None:
            if os.path.exists(AlignMonitorFile):
                command += ' -f "{0}"'.format(AlignMonitorFile)
        cmd = shlex.split(command)
        if self.autoUpdate:  # If we are not in autoUpdate mode, then we do not post to the Alignment Monitoring ELOG
            FNULL = open(os.devnull, 'w')
            subprocess.call(cmd, stdout=FNULL, stderr=subprocess.STDOUT)

        # Email
        subj2 = ''
        if self.testing:
            subj2 += 'TEST [' + self.dataVariant + ']: '
        text2 = "The Mirror "
        text2 += str(activity)
        text2 += " for RICH"
        text2 += str(self.whichRich)
        if conv:
            text2 += " ran for "
            text2 += alignment_time_str
            text2 += " and converged. "
            subj2 += 'RICH{0} alignment {1} converged: New ELOG entry'.format(
                self.whichRich, self.directoryTime)
        else:
            text2 += " ran for "
            text2 += alignment_time_str
            text2 += " and did not converge. Approximately "
            subj2 += 'RICH{0} alignment {1} incomplete: New ELOG entry'.format(
                self.whichRich, self.directoryTime)
        text2 += str(nEvents)
        text2 += " events per iteration were processed. "
        text2 += str(n_it + 1)
        text2 += " iteration(s) occured. \n"
        if DBUpdated is None:
            text2 += "\n There was a significant change from the alignment in the DB, or a change in magnet polarity. However the automatic update feature is off or was disabled during the alignment (The final state of autoUpdate was set to False), so the DB was not updated. If the automatic update feature was disabled, there may not have been enough events processed, or a sanity check was not satisfied. A MirrAlign expert should investigate whether this draft alignment (" + vN_DB_str + ") should be installed. \n"  # We should probably provide more info here
        elif DBUpdated:
            text2 += "\n Due to a significant change from the alignment in the DB, or a change in magnet polarity, the DB was automatically updated to " + vN_DB_str + ". \n"  # We should probably provide more info here
        else:
            text2 += "\n The alignment in the DB is still " + vN_DB_str + "."

        text2 += "\n More information is available at the following links: \n\n "
        text2 += 'https://lblogbook.cern.ch/RICH/?System=%5EAlignment%24&Subject={0}&mode=full'.format(
            self.directoryTime)
        text2 += '\n https://lbrich.cern.ch/alignmentview.php'
        text2 += '\n https://lbrich.cern.ch/plot/index.html?rich=' + str(
            self.whichRich) + '&source=mirror'

        text2 += "\n\n Fill number of first run processed: " + str(
            FillAlignment)

        if not self.autoUpdate:
            text2 += " \n\n The initial state of autoUpdate was set to False in the configuration."

        self.sendEmail(subj2, text2,
                       self.emails if not self.testing else self.experts)

        return messageID

    def writeSummary(self,
                     vN_DB_str,
                     conv,
                     n_it,
                     farm_config_time,
                     elapsed_runtimes,
                     summaryFile,
                     DBUpdated,
                     nEvents,
                     nEventsList,
                     flat=False):
        hh, r = divmod(farm_config_time, 3600)
        mm, ss = divmod(r, 60)
        farm_config_time_str = "{:0>2}:{:0>2}:{:05.2f}".format(
            int(hh), int(mm), ss)
        hh, r = divmod(elapsed_runtimes['elapsed_time_run'], 3600)
        mm, ss = divmod(r, 60)
        alignment_time_str = "{:0>2}:{:0>2}:{:05.2f}".format(
            int(hh), int(mm), ss)
        MirrCombinFit_time_str = str(
            elapsed_runtimes['elapsed_times_RichMirrCombinFit'])
        MirrAlign_time_str = str(
            elapsed_runtimes['elapsed_times_RichMirrAlign'])

        if flat is False:
            import OnlineEnv as Online
            runs = Online.DeferredRuns
        else:
            runs = []
        runsString = ''
        fillsString = ''
        polaritiesString = ''
        for i in range(0, len(runs)):
            runsString += str(runs[i]) + '  '
            fill = self.getFillNumber(runs[i])
            if fill is not None:
                fillsString += '  ' + str(fill) + '  '
            else:
                fillsString += '  ????  '
            polarity = self.getPolarity(runs[i])
            if polarity == 'UP':
                polaritiesString += '    UP  '
            elif polarity == 'DOWN':
                polaritiesString += '  DOWN  '
            elif polarity == 'OFF':
                polaritiesString += '   OFF  '
            else:  #something went wrong
                polaritiesString += '  ????  '

        if os.path.exists(summaryFile):
            os.remove(summaryFile)
        f = open(summaryFile, 'w')
        args = ''
        args += 'richDetector            = ' + str(self.whichRich) + '\n'
        args += 'DB updated              = ' + str(DBUpdated) + '\n'
        if DBUpdated is None:
            args += '                          (Alignment changed, but autoUpdate started as False or was disabled) \n'
        args += 'resulting DB            = ' + vN_DB_str + '\n'
        args += 'converged               = ' + str(conv) + '\n'
        args += 'Number of iterations    = ' + str(n_it + 1) + '\n'
        args += 'Final iteration label   = ' + '_i' + str(n_it) + '\n'
        args += 'Number of events        = ' + str(nEvents) + '\n'
        args += 'runs                    = ' + runsString + '\n'
        args += 'fills                   = ' + fillsString + '\n'
        args += 'polarities              = ' + polaritiesString + '\n'
        args += 'dataVariant             = ' + str(self.dataVariant) + '\n'
        args += 'savedir                 = ' + self.savedir + '\n'
        args += 'magnFactorsMode         = ' + str(self.magnFactorsMode) + '\n'
        args += 'magnifDir               = ' + str(self.magnifDir) + '\n'
        args += 'XML_YAML_configDir      = ' + str(
            self.XML_YAML_configDir) + '\n'
        args += 'xmlMDCS                 = ' + str(self.xmlMDCS) + '\n'
        args += 'combAndMirrSubsets      = ' + str(
            self.combAndMirrSubsets) + '\n'
        print("INFO: LHCbA config time (readable): " + farm_config_time_str)
        # print("INFO: Note, the LHCbA config time may be inaccurate if the alignment is run manually with LHCbA AutoPilot OFF.")
        args += 'LHCbA config time       = ' + farm_config_time_str + '\n'
        print("INFO: Alignment time    (readable): " + alignment_time_str)
        args += 'Alignment time          = ' + alignment_time_str + '\n'
        finalCKres = math.ceil(self.finalCKres * 10000) / 10000
        finalCKresErr = math.ceil(self.finalCKresErr * 10000) / 10000
        args += 'CK angle resolution     = ' + str(finalCKres) + '\n'
        args += '         error          = ' + str(finalCKresErr) + '\n'
        args += 'startXMLFile            = ' + str(
            self.alignConf.getProp('startXMLFile')) + '\n'
        args += 'compareXMLFile          = ' + str(
            self.alignConf.getProp('compareXMLFile')) + '\n'
        args += 'startYAMLFile           = ' + str(
            self.alignConf.getProp('startYAMLFile')) + '\n'
        args += 'compareYAMLFile         = ' + str(
            self.alignConf.getProp('compareYAMLFile')) + '\n'
        args += 'YAMLio                  = ' + str(self.YAMLio) + '\n'
        args += 'useDD4Hep               = ' + str(self.useDD4hep) + '\n'
        args += '\n'
        args += 'autoUpdate (initial)    = ' + str(self.autoUpdate) + '\n'
        args += 'autoUpdateNoV           = ' + str(self.autoUpdateNoV) + '\n'
        args += 'testing                 = ' + str(self.testing) + '\n'
        args += 'requiredEvents          = ' + str(
            self.alignConf.getProp('requiredEvents')) + '\n'
        args += 'minAverageBinPop        = ' + str(
            self.minAverageBinPop) + '\n'
        args += 'phiBinFactor            = ' + str(self.phiBinFactor) + '\n'
        args += 'minFracPhiBinsPopulated = ' + str(
            self.minFracPhiBinsPopulated) + '\n'
        args += 'deltaThetaWindow        = ' + str(
            self.deltaThetaWindow) + '\n'
        args += 'combinFitMethod         = ' + str(self.combinFitMethod) + '\n'
        args += 'solutionMethod          = ' + str(self.solutionMethod) + '\n'
        args += 'coeffCalibTilt          = ' + str(self.coeffCalibTilt) + '\n'
        args += 'fixSinusoidShift        = ' + str(
            self.fixSinusoidShift) + '\n'
        args += 'maxIterations           = ' + str(
            self.alignConf.getProp('maxIterations')) + '\n'
        args += 'EvtMax                  = ' + str(
            self.alignConf.getProp('EvtMax')) + '\n'
        args += 'regularizationMode      = ' + str(
            self.alignConf.getProp('regularizationMode')) + '\n'
        args += 'stopDecisionMode        = ' + str(
            self.alignConf.getProp('stopDecisionMode')) + '\n'
        args += 'stopToleranceMode       = ' + str(
            self.alignConf.getProp('stopToleranceMode')) + '\n'
        args += 'nominalResolutionSigma  = ' + str(
            self.alignConf.getProp('nominalResolutionSigma')) + '\n'
        args += 'stopSigmaFraction       = ' + str(
            self.alignConf.getProp('stopSigmaFraction')) + '\n'
        args += 'warningFactor           = ' + str(
            self.alignConf.getProp('warningFactor')) + '\n'
        args += 'displayMode             = ' + str(
            self.alignConf.getProp(
                'displayMode')) + '\n'  # not really being used yet!
        args += 'stopTolerancePriY       = ' + str(
            self.alignConf.getProp('stopTolerancePriY')) + '\n'
        args += 'stopTolerancePriZ       = ' + str(
            self.alignConf.getProp('stopTolerancePriZ')) + '\n'
        args += 'stopToleranceSecY       = ' + str(
            self.alignConf.getProp('stopToleranceSecY')) + '\n'
        args += 'stopToleranceSecZ       = ' + str(
            self.alignConf.getProp('stopToleranceSecZ')) + '\n'
        args += 'nEventsList             = ' + str(nEventsList) + '\n'
        args += 'CombinFit times         = ' + MirrCombinFit_time_str + '\n'
        args += 'MirrAlign times         = ' + MirrAlign_time_str + '\n'
        # not provided yet (not sure if there is need):
        # MajItStart and MinItStart
        # WorkDir, ItNrFile, thisCase, tiltNames, and original SaveDir in configuration
        # initial nameStr
        f.write(args)
        f.close()
        sys.stdout.flush()

        CKresTrendFile = "/group/rich/AlignmentFiles/Logging/Rich" + str(
            self.whichRich) + "_CKresTrend.txt"
        UpdateTrendFile = "/group/rich/AlignmentFiles/Logging/Rich" + str(
            self.whichRich) + "_UpdateTrend.txt"

        if flat is False:
            firstFill = self.getFillNumber(runs[0])
        else:
            firstFill = None
            firstPolarity = None
            firstCollisionInfo = None
        if firstFill is not None:
            firstFill_str = '' + str(firstFill) + ''
        else:
            firstFill_str = '????'
        if firstFill is not None:
            firstPolarity = self.getPolarity(runs[0])
        if firstPolarity == 'UP':
            firstPolarity_str = '  UP'
        elif firstPolarity == 'DOWN':
            firstPolarity_str = 'DOWN'
        elif firstPolarity == 'OFF':
            firstPolarity_str = ' OFF'
        else:  #something went wrong
            firstPolarity_str = '????'
        if firstFill is not None:
            firstCollisionInfo = self.getCollisionInfo(runs[0])
        if firstCollisionInfo is not None:
            firstCollisionInfo_str = '' + str(firstCollisionInfo) + ''
        else:
            firstCollisionInfo_str = '????'

        if not self.testing:
            with open(CKresTrendFile, "a") as myfile:
                myfile.write(firstFill_str + " " +
                             (self.directoryTime).replace("_", "T") + " " +
                             str(self.finalCKres / 1000) + " " +
                             str(self.finalCKresErr / 1000) + '\n')
            with open(UpdateTrendFile, "a") as myfile:
                myfile.write(firstFill_str + "\t" + self.directoryTime + "\t" +
                             vN_DB_str + "\t" + firstPolarity_str + "\t" +
                             firstCollisionInfo_str + '\n')
            if "maybe" in vN_DB_str:
                maybeSubj = "A _maybe version has been produced by the mirror alignment"
                maybeText = "A _maybe version has been produced by the mirror alignment. \n As you address this, also make sure to update the entry in /group/rich/AlignmentFiles/Logging/Rich" + str(
                    self.whichRich
                ) + "_UpdateTrend.txt, with the result (i.e. a proper vN, or the existing vN). Also you may wish to comment out the relevant line in /group/rich/AlignmentFiles/Logging/Rich" + str(
                    self.whichRich
                ) + "_CKresTrend.txt if you do not wish it to appear in the CK angle resolution trend plot; you could also copy the previous line's resolution, or just leave it if it is not a fluctuation."
                self.sendEmail(maybeSubj, maybeText, self.emails)
        else:
            with open(CKresTrendFile, "a") as myfile:
                myfile.write("# TEST " + firstFill_str + " " +
                             (self.directoryTime).replace("_", "T") + " " +
                             str(self.finalCKres / 1000) + " " +
                             str(self.finalCKresErr / 1000) + '\n')
            with open(UpdateTrendFile, "a") as myfile:
                myfile.write("# TEST " + firstFill_str + "\t" +
                             self.directoryTime + "\t" + vN_DB_str + "\t" +
                             firstPolarity_str + "\t" +
                             firstCollisionInfo_str + '\n')
            if "testing_maybe" in vN_DB_str:
                maybeSubj = "A _testing_maybe version has been produced by the mirror alignment"
                maybeText = "A _testing_maybe version has been produced by the mirror alignment. \n If you choose to use this alignment for this fill(s), make sure as you address this to update the entry in /group/rich/AlignmentFiles/Logging/Rich" + str(
                    self.whichRich
                ) + "_UpdateTrend.txt, with the result (i.e. a proper vN). Also uncomment the '# TEST ' part of the relevant line in /group/rich/AlignmentFiles/Logging/Rich" + str(
                    self.whichRich) + "_CKresTrend.txt please."
                self.sendEmail(maybeSubj, maybeText, self.experts)

    def sendEmail(self, subject, text, receiversList):
        import smtplib
        sender = 'lhcb-rich-mirror-alignment-development@cern.ch'
        message = 'From: MirrAlign Monitor <'
        message += sender
        message += '>\n'
        message += 'To:'
        for ie in receiversList:
            message += ' <'
            message += ie
            message += '>,'
        message = message[:-1]
        message += '\n'
        message += 'Subject: '
        message += subject
        message += '\n\n'
        message += '\n{0}\n\n'.format(text)
        try:
            smtpObj = smtplib.SMTP('localhost')
            smtpObj.sendmail(sender, receiversList, message)
            print("INFO: Successfully sent email")
        except smtplib.SMTPException:
            print("ERROR: Unable to send email")

    # Engineered from https://gitlab.cern.ch/lhcb/Lovell/blob/master/VeloAnalysisFramework/python/veloview/utils/rundb.py
    def getFillNumber(self, runNumber):
        response = self.queryRunDB(runNumber)
        # Check for null responses
        if not response:
            return None
        try:
            fill = response['fillid']
        except KeyError:
            # Fill ID not included in JSON response
            return None
        return fill

    # Engineered from https://gitlab.cern.ch/lhcb/Lovell/blob/master/VeloAnalysisFramework/python/veloview/utils/rundb.py
    def getCollisionInfo(self, runNumber):
        response = self.queryRunDB(runNumber)
        # Check for null responses
        if not response:
            return None
        try:
            collisionInfo = ""
            # OLD, doesn't work because hlt2TriggerSettings are not assigned in the RunDB until after the mirror alignment version is provided
            # collisionInfo = '|'.join(response['hlt2TriggerSettings'].split('|')[1:2])[7:] # This should extract the L0TCK recipe name, see https://lbtwiki.cern.ch/bin/view/L0/L0TCK for recipes
        except KeyError:
            # CollisionInfo not included in JSON response
            return None
        return collisionInfo

    # Engineered from https://gitlab.cern.ch/lhcb/Lovell/blob/master/VeloAnalysisFramework/python/veloview/utils/rundb.py
    def getPolarity(self, runNumber):
        # only possible polarity states: UP, DOWN, OFF
        response = self.queryRunDB(runNumber)
        # Check for null responses
        if not response:
            return None
        try:
            polarity = response['magnetState']
        except KeyError:
            # Polarity not included in JSON response
            return None
        # API returns UP for magnet up, DOWN for magnet down, OFF for off
        return polarity

    # Engineered from https://gitlab.cern.ch/lhcb/Lovell/blob/master/VeloAnalysisFramework/python/veloview/utils/rundb.py
    def queryRunDB(self, runNumber):
        import json, urllib
        RUNDB_API_URL = 'http://lbrundb.cern.ch/api/run/{0}/'  # e.g. see http://lbrundb.cern.ch/api/run/185380/ for what's available
        request = urllib.urlopen(RUNDB_API_URL.format(runNumber))
        # Make sure the request was OK, not e.g. 404 not found
        if request.getcode() != 200:
            return None
        try:
            run_json = json.load(request)
        except ValueError:
            # Couldn't decode the response
            return None
        return run_json

    def getNEventsListList(self, n_it):
        yaml = YAML()
        with open(self.workdir + '/' + 'rich' + str(self.whichRich) +
                  '_variant.yml') as finp:
            current_variant = yaml.load(finp)
        self.fileprename = self.workdir + current_variant + "_histos_i"
        nEventsList = []
        filename = []
        thisFile = []
        hist = []
        for j in range(0, n_it + 1):
            filename.append(self.fileprename + str(j) + '.root')
            thisFile.append(TFile(filename[j], 'read'))
            hist.append(thisFile[j].Get(
                'RICH/RichODIN/TriggerType'))  # won't work on MC
            nEventsList.append(int(hist[j].GetEntries()))
        nEvents = int(
            round(float(sum(nEventsList)) / float(n_it + 1))
        )  # If mag factors float, this only reports for the data with untilted XML/YAML
        return [nEventsList, nEvents]

    def quickWriteInLogbook(self, subject, text, edit=None):
        host = 'lblogbook.cern.ch'
        port = 8080
        username = 'common Common\\!'
        logbook = 'RICH'
        author = 'MirrAlign Monitor'  # Or any other name
        activity = 'Alignment'
        instruction_file = 'https://twiki.cern.ch/twiki/bin/view/LHCb/LHCbRichMirrorAlignShiftInfo'

        text += " \n Shifter's instructions can be found at "
        text += str(instruction_file)
        text += "."

        command = 'elog -h "'
        command += str(host)
        command += '" -p '
        command += str(port)
        try:
            command += ' -e ' + str(int(edit))
        except:
            pass
        command += ' -l "'
        command += str(logbook)
        command += '" -u '
        command += str(username)
        command += ' -a "Author='
        command += str(author)
        #command += '" -a "System='
        command += '" -a "System_19='  # Change 19 if Alignment changes its order in the ELOG (ask the current RICH expert)
        command += str(activity)
        command += '" -a "Subject='
        command += str(subject)
        command += '" "'
        command += str(text)
        command += '"'
        cmd = shlex.split(command)
        try:
            elogret = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            messageID = int(
                re.search('.*Message successfully transmitted, ID=(\d*)',
                          elogret).groups()[0])
        except:
            messageID = None

        return messageID
